@@ -1,15 +1,23 @@
 """
+The `Document` that implements all the text operations/querying.
 """
 from __future__ import unicode_literals
 
+import bisect
 import re
-import six
 import string
+import weakref
+
+import six
+from six.moves import map, range
 
-from .selection import SelectionType, SelectionState
 from .clipboard import ClipboardData
+from .filters import vi_mode
+from .selection import PasteMode, SelectionState, SelectionType
 
-__all__ = ('Document',)
+__all__ = [
+    'Document',
+]
 
 
 # Regex for finding "words" in documents. (We consider a group of alnum
@@ -26,6 +34,40 @@ _FIND_BIG_WORD_RE = re.compile(r'([^\s]+)')
 _FIND_CURRENT_BIG_WORD_RE = re.compile(r'^([^\s]+)')
 _FIND_CURRENT_BIG_WORD_INCLUDE_TRAILING_WHITESPACE_RE = re.compile(r'^([^\s]+\s*)')
 
+# Share the Document._cache between all Document instances.
+# (Document instances are considered immutable. That means that if another
+# `Document` is constructed with the same text, it should have the same
+# `_DocumentCache`.)
+_text_to_document_cache = weakref.WeakValueDictionary()  # Maps document.text to DocumentCache instance.
+
+
+class _ImmutableLineList(list):
+    """
+    Some protection for our 'lines' list, which is assumed to be immutable in the cache.
+    (Useful for detecting obvious bugs.)
+    """
+    def _error(self, *a, **kw):
+        raise NotImplementedError('Attempt to modify an immutable list.')
+
+    __setitem__ = _error
+    append = _error
+    clear = _error
+    extend = _error
+    insert = _error
+    pop = _error
+    remove = _error
+    reverse = _error
+    sort = _error
+
+
+class _DocumentCache(object):
+    def __init__(self):
+        #: List of lines for the Document text.
+        self.lines = None
+
+        #: List of index positions, pointing to the start of all the lines.
+        self.line_indexes = None
+
 
 class Document(object):
     """
@@ -39,7 +81,7 @@ class Document(object):
     :param cursor_position: int
     :param selection: :class:`.SelectionState`
     """
-    __slots__ = ('text', 'cursor_position', 'selection', '_lines_cache')
+    __slots__ = ('_text', '_cursor_position', '_selection', '_cache')
 
     def __init__(self, text='', cursor_position=None, selection=None):
         assert isinstance(text, six.text_type), 'Got %r' % text
@@ -56,14 +98,51 @@ class Document(object):
         if cursor_position is None:
             cursor_position = len(text)
 
-        self.text = text
-        self.cursor_position = cursor_position
-        self.selection = selection
-        self._lines_cache = None
+        # Keep these attributes private. A `Document` really has to be
+        # considered to be immutable, because otherwise the caching will break
+        # things. Because of that, we wrap these into read-only properties.
+        self._text = text
+        self._cursor_position = cursor_position
+        self._selection = selection
+
+        # Cache for lines/indexes. (Shared with other Document instances that
+        # contain the same text.
+        try:
+            self._cache = _text_to_document_cache[self.text]
+        except KeyError:
+            self._cache = _DocumentCache()
+            _text_to_document_cache[self.text] = self._cache
+
+        # XX: For some reason, above, we can't use 'WeakValueDictionary.setdefault'.
+        #     This fails in Pypy3. `self._cache` becomes None, because that's what
+        #     'setdefault' returns.
+        # self._cache = _text_to_document_cache.setdefault(self.text, _DocumentCache())
+        # assert self._cache
 
     def __repr__(self):
         return '%s(%r, %r)' % (self.__class__.__name__, self.text, self.cursor_position)
 
+    def __eq__(self, other):
+        assert isinstance(other, Document)
+        return (self.text == other.text and
+                self.cursor_position == other.cursor_position and
+                self.selection == other.selection)
+
+    @property
+    def text(self):
+        " The document text. "
+        return self._text
+
+    @property
+    def cursor_position(self):
+        " The document cursor position. "
+        return self._cursor_position
+
+    @property
+    def selection(self):
+        " :class:`.SelectionState` object. "
+        return self._selection
+
     @property
     def current_char(self):
         """ Return character under cursor or an empty string. """
@@ -85,21 +164,53 @@ class Document(object):
     @property
     def current_line_before_cursor(self):
         """ Text from the start of the line until the cursor. """
-        return self.text_before_cursor.split('\n')[-1]
+        _, _, text = self.text_before_cursor.rpartition('\n')
+        return text
 
     @property
     def current_line_after_cursor(self):
         """ Text from the cursor until the end of the line. """
-        return self.text_after_cursor.split('\n')[0]
+        text, _, _ = self.text_after_cursor.partition('\n')
+        return text
 
     @property
     def lines(self):
-        " Array of all the lines. "
+        """
+        Array of all the lines.
+        """
         # Cache, because this one is reused very often.
-        if self._lines_cache is None:
-            self._lines_cache = self.text.split('\n')
+        if self._cache.lines is None:
+            self._cache.lines = _ImmutableLineList(self.text.split('\n'))
 
-        return self._lines_cache
+        return self._cache.lines
+
+    @property
+    def _line_start_indexes(self):
+        """
+        Array pointing to the start indexes of all the lines.
+        """
+        # Cache, because this is often reused. (If it is used, it's often used
+        # many times. And this has to be fast for editing big documents!)
+        if self._cache.line_indexes is None:
+            # Create list of line lengths.
+            line_lengths = map(len, self.lines)
+
+            # Calculate cumulative sums.
+            indexes = [0]
+            append = indexes.append
+            pos = 0
+
+            for line_length in line_lengths:
+                pos += line_length + 1
+                append(pos)
+
+            # Remove the last item. (This is not a new line.)
+            if len(indexes) > 1:
+                indexes.pop()
+
+            self._cache.line_indexes = indexes
+
+        return self._cache.line_indexes
 
     @property
     def lines_from_current(self):
@@ -155,24 +266,40 @@ class Document(object):
         """
         Current row. (0-based.)
         """
-        return len(self.text_before_cursor.split('\n')) - 1
+        row, _ = self._find_line_start_index(self.cursor_position)
+        return row
 
     @property
     def cursor_position_col(self):
         """
         Current column. (0-based.)
         """
-        return len(self.current_line_before_cursor)
+        # (Don't use self.text_before_cursor to calculate this. Creating
+        # substrings and doing rsplit is too expensive for getting the cursor
+        # position.)
+        _, line_start_index = self._find_line_start_index(self.cursor_position)
+        return self.cursor_position - line_start_index
 
-    def translate_index_to_position(self, index):  # TODO: make this 0-based indexed!!!
+    def _find_line_start_index(self, index):
+        """
+        For the index of a character at a certain line, calculate the index of
+        the first character on that line.
+
+        Return (row, index) tuple.
+        """
+        indexes = self._line_start_indexes
+
+        pos = bisect.bisect_right(indexes, index) - 1
+        return pos, indexes[pos]
+
+    def translate_index_to_position(self, index):
         """
         Given an index for the text, return the corresponding (row, col) tuple.
         (0-based. Returns (0, 0) for index=0.)
         """
-        text_before_position = self.text[:index]
-
-        row = text_before_position.count('\n')
-        col = len(text_before_position.split('\n')[-1])
+        # Find start of this line.
+        row, row_index = self._find_line_start_index(index)
+        col = index - row_index
 
         return row, col
 
@@ -180,8 +307,21 @@ class Document(object):
         """
         Given a (row, col) tuple, return the corresponding index.
         (Row and col params are 0-based.)
+
+        Negative row/col values are turned into zero.
         """
-        result = len('\n'.join(self.lines[:row])) + (len('\n') if row > 0 else 0) + col
+        try:
+            result = self._line_start_indexes[row]
+            line = self.lines[row]
+        except IndexError:
+            if row < 0:
+                result = self._line_start_indexes[0]
+                line = self.lines[0]
+            else:
+                result = self._line_start_indexes[-1]
+                line = self.lines[-1]
+
+        result += max(0, min(col, len(line)))
 
         # Keep in range. (len(self.text) is included, because the cursor can be
         # right after the end of the text as well.)
@@ -196,21 +336,21 @@ class Document(object):
     @property
     def is_cursor_at_the_end_of_line(self):
         """ True when the cursor is at the end of this line. """
-        return self.cursor_position_col == len(self.current_line)
+        return self.current_char in ('\n', '')
 
     def has_match_at_current_position(self, sub):
         """
         `True` when this substring is found at the cursor position.
         """
-        return self.text[self.cursor_position:].find(sub) == 0
+        return self.text.find(sub, self.cursor_position) == self.cursor_position
 
     def find(self, sub, in_current_line=False, include_current_position=False,
-            ignore_case=False, count=1):
+             ignore_case=False, count=1):
         """
         Find `text` after the cursor, return position relative to the cursor
         position. Return `None` if nothing was found.
 
-        :param count: Find the n-th occurance.
+        :param count: Find the n-th occurrence.
         """
         assert isinstance(ignore_case, bool)
 
@@ -240,7 +380,7 @@ class Document(object):
 
     def find_all(self, sub, ignore_case=False):
         """
-        Find all occurances of the substring. Return a list of absolute
+        Find all occurrences of the substring. Return a list of absolute
         positions in the document.
         """
         flags = re.IGNORECASE if ignore_case else 0
@@ -251,7 +391,7 @@ class Document(object):
         Find `text` before the cursor, return position relative to the cursor
         position. Return `None` if nothing was found.
 
-        :param count: Find the n-th occurance.
+        :param count: Find the n-th occurrence.
         """
         if in_current_line:
             before_cursor = self.current_line_before_cursor[::-1]
@@ -268,32 +408,56 @@ class Document(object):
         except StopIteration:
             pass
 
-    def get_word_before_cursor(self, WORD=False):
+    def get_word_before_cursor(self, WORD=False, pattern=None):
         """
         Give the word before the cursor.
         If we have whitespace before the cursor this returns an empty string.
-        """
-        if self.text_before_cursor[-1:].isspace():
-            return ''
-        else:
-            return self.text_before_cursor[self.find_start_of_previous_word(WORD=WORD):]
 
-    def find_start_of_previous_word(self, count=1, WORD=False):
+        :param pattern: (None or compiled regex). When given, use this regex
+            pattern.
+        """
+        if self._is_word_before_cursor_complete(WORD=WORD, pattern=pattern):
+            # Space before the cursor or no text before cursor.
+            return ''
+
+        text_before_cursor = self.text_before_cursor
+        start = self.find_start_of_previous_word(WORD=WORD, pattern=pattern)
+
+        return text_before_cursor[len(text_before_cursor) + start:]
+
+    def _is_word_before_cursor_complete(self, WORD=False, pattern=None):
+        if pattern:
+            return self.find_start_of_previous_word(pattern=pattern) is None
+        else:
+            return self.text_before_cursor == '' or self.text_before_cursor[-1:].isspace()
+
+    def find_start_of_previous_word(self, count=1, WORD=False, pattern=None):
         """
         Return an index relative to the cursor position pointing to the start
         of the previous word. Return `None` if nothing was found.
+
+        :param pattern: (None or compiled regex). When given, use this regex
+            pattern.
         """
+        assert not (WORD and pattern)
+
         # Reverse the text before the cursor, in order to do an efficient
         # backwards search.
         text_before_cursor = self.text_before_cursor[::-1]
 
-        regex = _FIND_BIG_WORD_RE if WORD else _FIND_WORD_RE
+        if pattern:
+            regex = pattern
+        elif WORD:
+            regex = _FIND_BIG_WORD_RE
+        else:
+            regex = _FIND_WORD_RE
+
         iterator = regex.finditer(text_before_cursor)
 
         try:
             for i, match in enumerate(iterator):
                 if i + 1 == count:
-                    return - match.end(1)
+                    return - match.end(0)
         except StopIteration:
             pass
 
@@ -349,6 +513,9 @@ class Document(object):
         Return an index relative to the cursor position pointing to the start
         of the next word. Return `None` if nothing was found.
         """
+        if count < 0:
+            return self.find_previous_word_beginning(count=-count, WORD=WORD)
+
         regex = _FIND_BIG_WORD_RE if WORD else _FIND_WORD_RE
         iterator = regex.finditer(self.text_after_cursor)
 
@@ -368,6 +535,9 @@ class Document(object):
         Return an index relative to the cursor position pointing to the end
         of the next word. Return `None` if nothing was found.
         """
+        if count < 0:
+            return self.find_previous_word_ending(count=-count, WORD=WORD)
+
         if include_current_position:
             text = self.text_after_cursor
         else:
@@ -392,8 +562,11 @@ class Document(object):
     def find_previous_word_beginning(self, count=1, WORD=False):
         """
         Return an index relative to the cursor position pointing to the start
-        of the next word. Return `None` if nothing was found.
+        of the previous word. Return `None` if nothing was found.
         """
+        if count < 0:
+            return self.find_next_word_beginning(count=-count, WORD=WORD)
+
         regex = _FIND_BIG_WORD_RE if WORD else _FIND_WORD_RE
         iterator = regex.finditer(self.text_before_cursor[::-1])
 
@@ -404,6 +577,30 @@ class Document(object):
         except StopIteration:
             pass
 
+    def find_previous_word_ending(self, count=1, WORD=False):
+        """
+        Return an index relative to the cursor position pointing to the end
+        of the previous word. Return `None` if nothing was found.
+        """
+        if count < 0:
+            return self.find_next_word_ending(count=-count, WORD=WORD)
+
+        text_before_cursor = self.text_after_cursor[:1] + self.text_before_cursor[::-1]
+
+        regex = _FIND_BIG_WORD_RE if WORD else _FIND_WORD_RE
+        iterator = regex.finditer(text_before_cursor)
+
+        try:
+            for i, match in enumerate(iterator):
+                # Take first match, unless it's the word on which we're right now.
+                if i == 0 and match.start(1) == 0:
+                    count += 1
+
+                if i + 1 == count:
+                    return -match.start(1) + 1
+        except StopIteration:
+            pass
+
     def find_next_matching_line(self, match_func, count=1):
         """
         Look downwards for empty lines.
@@ -442,94 +639,119 @@ class Document(object):
         """
         Relative position for cursor left.
         """
+        if count < 0:
+            return self.get_cursor_right_position(-count)
+
         return - min(self.cursor_position_col, count)
 
     def get_cursor_right_position(self, count=1):
         """
         Relative position for cursor_right.
         """
+        if count < 0:
+            return self.get_cursor_left_position(-count)
+
         return min(count, len(self.current_line_after_cursor))
 
-    def get_cursor_up_position(self, count=1):
+    def get_cursor_up_position(self, count=1, preferred_column=None):
         """
         Return the relative cursor position (character index) where we would be if the
         user pressed the arrow-up button.
+
+        :param preferred_column: When given, go to this column instead of
+                                 staying at the current column.
         """
         assert count >= 1
+        column = self.cursor_position_col if preferred_column is None else preferred_column
 
-        count = min(self.text_before_cursor.count('\n'), count)
+        return self.translate_row_col_to_index(
+            max(0, self.cursor_position_row - count), column) - self.cursor_position
 
-        if count:
-            pos = self.cursor_position_col
-
-            lines = self.text_before_cursor.split('\n')
-            skip_lines = '\n'.join(lines[-count-1:])
-            new_line = lines[-count-1]
-
-            # When the current line is longer then the previous, move to the
-            # last character of the previous line.
-            if pos > len(new_line):
-                return - len(skip_lines) + len(new_line)
-
-            # Otherwise find the corresponding position in the previous line.
-            else:
-                return - len(skip_lines) + pos
-        return 0
-
-    def get_cursor_down_position(self, count=1):
+    def get_cursor_down_position(self, count=1, preferred_column=None):
         """
         Return the relative cursor position (character index) where we would be if the
         user pressed the arrow-down button.
+
+        :param preferred_column: When given, go to this column instead of
+                                 staying at the current column.
         """
         assert count >= 1
+        column = self.cursor_position_col if preferred_column is None else preferred_column
 
-        count = min(self.text_after_cursor.count('\n'), count)
+        return self.translate_row_col_to_index(
+            self.cursor_position_row + count, column) - self.cursor_position
 
-        if count:
-            pos = self.cursor_position_col
-
-            lines = self.text_after_cursor.split('\n')
-            skip_lines = '\n'.join(lines[:count])
-            new_line = lines[count]
-
-            # When the current line is longer then the previous, move to the
-            # last character of the next line.
-            if pos > len(new_line):
-                return len(skip_lines) + len(new_line) + 1
-
-            # Otherwise find the corresponding position in the next line.
-            else:
-                return len(skip_lines) + pos + 1
-
-        return 0
-
-    @property
-    def matching_bracket_position(self):
+    def find_enclosing_bracket_right(self, left_ch, right_ch, end_pos=None):
         """
-        Return relative cursor position of matching [, (, { or < bracket.
+        Find the right bracket enclosing current position. Return the relative
+        position to the cursor position.
+
+        When `end_pos` is given, don't look past the position.
         """
+        if self.current_char == right_ch:
+            return 0
+
+        if end_pos is None:
+            end_pos = len(self.text)
+        else:
+            end_pos = min(len(self.text), end_pos)
+
         stack = 1
 
+        # Look forward.
+        for i in range(self.cursor_position + 1, end_pos):
+            c = self.text[i]
+
+            if c == left_ch:
+                stack += 1
+            elif c == right_ch:
+                stack -= 1
+
+            if stack == 0:
+                return i - self.cursor_position
+
+    def find_enclosing_bracket_left(self, left_ch, right_ch, start_pos=None):
+        """
+        Find the left bracket enclosing current position. Return the relative
+        position to the cursor position.
+
+        When `start_pos` is given, don't look past the position.
+        """
+        if self.current_char == left_ch:
+            return 0
+
+        if start_pos is None:
+            start_pos = 0
+        else:
+            start_pos = max(0, start_pos)
+
+        stack = 1
+
+        # Look backward.
+        for i in range(self.cursor_position - 1, start_pos - 1, -1):
+            c = self.text[i]
+
+            if c == right_ch:
+                stack += 1
+            elif c == left_ch:
+                stack -= 1
+
+            if stack == 0:
+                return i - self.cursor_position
+
+    def find_matching_bracket_position(self, start_pos=None, end_pos=None):
+        """
+        Return relative cursor position of matching [, (, { or < bracket.
+
+        When `start_pos` or `end_pos` are given. Don't look past the positions.
+        """
+
+        # Look for a match.
         for A, B in '()', '[]', '{}', '<>':
             if self.current_char == A:
-                for i, c in enumerate(self.text_after_cursor[1:]):
-                    if c == A:
-                        stack += 1
-                    elif c == B:
-                        stack -= 1
-
-                    if stack == 0:
-                        return i + 1
-
+                return self.find_enclosing_bracket_right(A, B, end_pos=end_pos) or 0
             elif self.current_char == B:
-                for i, c in enumerate(reversed(self.text_before_cursor)):
-                    if c == B:
-                        stack += 1
-                    elif c == A:
-                        stack -= 1
-
-                    if stack == 0:
-                        return - (i + 1)
+                return self.find_enclosing_bracket_left(A, B, start_pos=start_pos) or 0
 
         return 0
 
@@ -557,7 +779,7 @@ class Document(object):
         """
         Relative position for the last non blank character of this line.
         """
-        return len(self.current_line_after_cursor.rstrip())
+        return len(self.current_line.rstrip()) - self.cursor_position_col - 1
 
     def get_column_cursor_position(self, column):
         """
@@ -571,7 +793,7 @@ class Document(object):
 
         return column - current_column
 
-    def selection_range(self):
+    def selection_range(self):  # XXX: shouldn't this return `None` if there is no selection???
         """
         Return (from, to) tuple of the selection.
         start and end position are included.
@@ -588,11 +810,12 @@ class Document(object):
 
     def selection_ranges(self):
         """
-        Return a list of (from, to) tuples for the selection or none if nothing
-        was selected.  start and end position are always included in the
-        selection.
+        Return a list of `(from, to)` tuples for the selection or none if
+        nothing was selected. The upper boundary is not included.
 
         This will yield several (from, to) tuples in case of a BLOCK selection.
+        This will return zero ranges, like (8,8) for empty lines in a block
+        selection.
         """
         if self.selection:
             from_, to = sorted([self.cursor_position, self.selection.original_cursor_position])
@@ -603,23 +826,79 @@ class Document(object):
                 from_column, to_column = sorted([from_column, to_column])
                 lines = self.lines
 
+                if vi_mode():
+                    to_column += 1
+
                 for l in range(from_line, to_line + 1):
                     line_length = len(lines[l])
-                    if from_column < line_length:
+
+                    if from_column <= line_length:
                         yield (self.translate_row_col_to_index(l, from_column),
-                               self.translate_row_col_to_index(l, min(line_length - 1, to_column)))
+                               self.translate_row_col_to_index(l, min(line_length, to_column)))
             else:
                 # In case of a LINES selection, go to the start/end of the lines.
                 if self.selection.type == SelectionType.LINES:
-                    from_ = max(0, self.text[:from_].rfind('\n') + 1)
+                    from_ = max(0, self.text.rfind('\n', 0, from_) + 1)
 
-                    if self.text[to:].find('\n') >= 0:
-                        to += self.text[to:].find('\n')
+                    if self.text.find('\n', to) >= 0:
+                        to = self.text.find('\n', to)
                     else:
-                        to = len(self.text)
+                        to = len(self.text) - 1
+
+                # In Vi mode, the upper boundary is always included. For Emacs,
+                # that's not the case.
+                if vi_mode():
+                    to += 1
 
                 yield from_, to
 
+    def selection_range_at_line(self, row):
+        """
+        If the selection spans a portion of the given line, return a (from, to) tuple.
+
+        The returned upper boundary is not included in the selection, so
+        `(0, 0)` is an empty selection.  `(0, 1)`, is a one character selection.
+
+        Returns None if the selection doesn't cover this line at all.
+        """
+        if self.selection:
+            line = self.lines[row]
+
+            row_start = self.translate_row_col_to_index(row, 0)
+            row_end = self.translate_row_col_to_index(row, len(line))
+
+            from_, to = sorted([self.cursor_position, self.selection.original_cursor_position])
+
+            # Take the intersection of the current line and the selection.
+            intersection_start = max(row_start, from_)
+            intersection_end = min(row_end, to)
+
+            if intersection_start <= intersection_end:
+                if self.selection.type == SelectionType.LINES:
+                    intersection_start = row_start
+                    intersection_end = row_end
+
+                elif self.selection.type == SelectionType.BLOCK:
+                    _, col1 = self.translate_index_to_position(from_)
+                    _, col2 = self.translate_index_to_position(to)
+                    col1, col2 = sorted([col1, col2])
+
+                    if col1 > len(line):
+                        return  # Block selection doesn't cross this line.
+
+                    intersection_start = self.translate_row_col_to_index(row, col1)
+                    intersection_end = self.translate_row_col_to_index(row, col2)
+
+                _, from_column = self.translate_index_to_position(intersection_start)
+                _, to_column = self.translate_index_to_position(intersection_end)
+
+                # In Vi mode, the upper boundary is always included. For Emacs
+                # mode, that's not the case.
+                if vi_mode():
+                    to_column += 1
+
+                return from_column, to_column
+
     def cut_selection(self):
         """
         Return a (:class:`.Document`, :class:`.ClipboardData`) tuple, where the
@@ -637,8 +916,8 @@ class Document(object):
                     new_cursor_position = from_
 
                 remaining_parts.append(self.text[last_to:from_])
-                cut_parts.append(self.text[from_:to + 1])
-                last_to = to + 1
+                cut_parts.append(self.text[from_:to])
+                last_to = to
 
             remaining_parts.append(self.text[last_to:])
 
@@ -654,24 +933,30 @@ class Document(object):
         else:
             return self, ClipboardData('')
 
-    def paste_clipboard_data(self, data, before=False, count=1):
+    def paste_clipboard_data(self, data, paste_mode=PasteMode.EMACS, count=1):
         """
         Return a new :class:`.Document` instance which contains the result if
         we would paste this data at the current cursor position.
 
-        :param before: Paste before the cursor position. (For line/character mode.)
+        :param paste_mode: Where to paste. (Before/after/emacs.)
         :param count: When >1, Paste multiple times.
         """
         assert isinstance(data, ClipboardData)
+        assert paste_mode in (PasteMode.VI_BEFORE, PasteMode.VI_AFTER, PasteMode.EMACS)
+
+        before = (paste_mode == PasteMode.VI_BEFORE)
+        after = (paste_mode == PasteMode.VI_AFTER)
 
         if data.type == SelectionType.CHARACTERS:
-            if before:
-                new_text = self.text_before_cursor + data.text * count + self.text_after_cursor
-                new_cursor_position = self.cursor_position + len(data.text) * count - 1
-            else:
+            if after:
                 new_text = (self.text[:self.cursor_position + 1] + data.text * count +
                             self.text[self.cursor_position + 1:])
-                new_cursor_position = self.cursor_position + len(data.text) * count
+            else:
+                new_text = self.text_before_cursor + data.text * count + self.text_after_cursor
+
+            new_cursor_position = self.cursor_position + len(data.text) * count
+            if before:
+                new_cursor_position -= 1
 
         elif data.type == SelectionType.LINES:
             l = self.cursor_position_row
@@ -685,7 +970,7 @@ class Document(object):
                 new_text = '\n'.join(lines)
 
         elif data.type == SelectionType.BLOCK:
-            lines = self.lines
+            lines = self.lines[:]
             start_line = self.cursor_position_row
             start_column = self.cursor_position_col + (0 if before else 1)
 
@@ -715,6 +1000,36 @@ class Document(object):
 
         return count
 
+    def start_of_paragraph(self, count=1, before=False):
+        """
+        Return the start of the current paragraph. (Relative cursor position.)
+        """
+        def match_func(text):
+            return not text or text.isspace()
+
+        line_index = self.find_previous_matching_line(match_func=match_func, count=count)
+
+        if line_index:
+            add = 0 if before else 1
+            return min(0, self.get_cursor_up_position(count=-line_index) + add)
+        else:
+            return -self.cursor_position
+
+    def end_of_paragraph(self, count=1, after=False):
+        """
+        Return the end of the current paragraph. (Relative cursor position.)
+        """
+        def match_func(text):
+            return not text or text.isspace()
+
+        line_index = self.find_next_matching_line(match_func=match_func, count=count)
+
+        if line_index:
+            add = 0 if after else 1
+            return max(0, self.get_cursor_down_position(count=line_index) - add)
+        else:
+            return len(self.text_after_cursor)
+
     # Modifiers.
 
     def insert_after(self, text):
