@@ -3,10 +3,11 @@ Data structures for the selection.
 """
 from __future__ import unicode_literals
 
-__all__ = (
+__all__ = [
     'SelectionType',
+    'PasteMode',
     'SelectionState',
-)
+]
 
 
 class SelectionType(object):
@@ -23,6 +24,12 @@ class SelectionType(object):
     BLOCK = 'BLOCK'
 
 
+class PasteMode(object):
+    EMACS = 'EMACS'  # Yank like emacs.
+    VI_AFTER = 'VI_AFTER'  # When pressing 'p' in Vi.
+    VI_BEFORE = 'VI_BEFORE'  # When pressing 'P' in Vi.
+
+
 class SelectionState(object):
     """
     State of the current selection.
